package io.vertx.config.kubernetes;

import io.vertx.config.spi.ConfigStore;
import io.vertx.config.spi.utils.JsonObjectHelper;
import io.vertx.core.AsyncResult;
import io.vertx.core.Future;
import io.vertx.core.Handler;
import io.vertx.core.Vertx;
import io.vertx.core.buffer.Buffer;
import io.vertx.core.json.JsonObject;
import io.vertx.ext.web.client.HttpResponse;
import io.vertx.ext.web.client.WebClient;
import io.vertx.ext.web.client.WebClientOptions;

import java.util.HashMap;
import java.util.Map;
import java.util.Objects;
import java.util.stream.Collectors;

/**
 * An implementation of configuration store reading config map from Kubernetes.
 */
public class ConfigMapStore implements ConfigStore {
  private static final String KUBERNETES_NAMESPACE = System.getenv("KUBERNETES_NAMESPACE");
  private final Vertx vertx;
  private final JsonObject configuration;
  private final String namespace;
  private final String name;
  private final String key;
  private final boolean secret;
  private final boolean optional;

  private final WebClient client;
  private String token;


  public ConfigMapStore(Vertx vertx, JsonObject configuration) {
    this.vertx = vertx;
    this.configuration = configuration;
    String ns = configuration.getString("namespace");
    if (ns == null) {
      if (KUBERNETES_NAMESPACE != null) {
        ns = KUBERNETES_NAMESPACE;
      } else {
        ns = "default";
      }
    }
    this.optional = configuration.getBoolean("optional", true);
    this.namespace = ns;
    this.name = configuration.getString("name");
    this.key = configuration.getString("key");
    this.secret = configuration.getBoolean("secret", false);
    int port = configuration.getInteger("port", 0);
    if (port == 0) {
      if (configuration.getBoolean("ssl", true)) {
        port = 443;
      } else {
        port = 80;
      }
    }

    String p = System.getenv("KUBERNETES_SERVICE_PORT");
    if (p != null) {
      port = Integer.valueOf(p);
    }

    String host = configuration.getString("host");
    String h = System.getenv("KUBERNETES_SERVICE_HOST");
    if (h != null) {
      host = h;
    }

    client = WebClient.create(vertx,
      new WebClientOptions()
        .setTrustAll(true)
        .setSsl(configuration.getBoolean("ssl", true))
        .setDefaultHost(host)
        .setDefaultPort(port)
        .setFollowRedirects(true)
    );

    Objects.requireNonNull(this.name);
  }

  @Override
  public synchronized void close(Handler<Void> completionHandler) {
    if (client != null) {
      client.close();
    }
  }


  private Future<String> getToken() {
    Future<String> result = Future.future();

    String token = configuration.getString("token");
    if (token != null && !token.trim().isEmpty()) {
      this.token = token;
      result.complete(token);
      return result;
    }

    // Read from file
    vertx.fileSystem().readFile(KubernetesUtils.OPENSHIFT_KUBERNETES_TOKEN_FILE, ar -> {
      if (ar.failed()) {
        result.tryFail(ar.cause());
      } else {
        this.token = ar.result().toString();
        result.tryComplete(ar.result().toString());
      }
    });

    return result;
  }

  @Override
  public synchronized void get(Handler<AsyncResult<Buffer>> completionHandler) {
    Future<String> retrieveToken;
    if (token == null) {
      retrieveToken = getToken();
    } else {
      retrieveToken = Future.succeededFuture(token);
    }

    retrieveToken
      .compose(token -> {
        String path = "/api/v1/namespaces/" + namespace;
        if (secret) {
          path += "/secrets/" + name;
        } else {
          path += "/configmaps/" + name;
        }

        Future<Buffer> future = Future.future();
        client.get(path)
          .putHeader("Authorization", "Bearer " + token)
          .send(ar -> {
            if (ar.failed()) {
              completionHandler.handle(ar.mapEmpty());
              return;
            }
            HttpResponse<Buffer> response = ar.result();
            if (response.statusCode() == 404) {
              if (optional) {
                future.complete(Buffer.buffer("{}"));
              } else {
                future.fail("Cannot find the config map '" + name + "' in '" + namespace + "'");
              }
            } else if (response.statusCode() == 403) {
              completionHandler.handle(Future.failedFuture("Access denied to configmap or secret in namespace "
                + namespace + ": " + name));
            } else if (response.statusCode() != 200) {
              if (optional) {
                future.complete(Buffer.buffer("{}"));
              } else {
                completionHandler.handle(Future.failedFuture("Cannot retrieve the configmap or secret in namespace "
                  + namespace + ": " + name + ", status code: " + response.statusCode() + ", error: "
                  + response.bodyAsString()));
              }
            } else {
              JsonObject data = response.bodyAsJsonObject().getJsonObject("data");
              if (data == null) {
                future.fail("Invalid secret of configmap in namespace " + namespace + " " + name + ", the data " +
                  "entry is empty");
                return;
              }
              if (this.key == null) {
                future.complete(new JsonObject(asObjectMap(data.getMap())).toBuffer());
              } else {
                String string = data.getString(this.key);
                if (string == null) {
                  future.fail("Cannot find key '" + this.key + "' in the configmap or secret '" + this.name + "'");
                } else {
                  future.complete(Buffer.buffer(string));
                }
              }
            }
          });
        return future;
      }).setHandler(completionHandler);
  }

  private static Map<String, Object> asObjectMap(Map<String, Object> source) {
    if (source == null) {
      return new HashMap<>();
    }
    return source.entrySet().stream().collect(Collectors.toMap(Map.Entry::getKey,
      entry -> JsonObjectHelper.convert(entry.getValue().toString())));
  }

}
